"""
parse.py: creates parse tree.
"""

from random import randrange

from .errors import InvalidMemLoc, InvalidInstruction
from .errors import UnknownName, InvalidDataType, InvalidSection
from .errors import InvalidArgument, MissingData, InvalidDataVal, MissingComma
from .errors import MissingOpenParen, MissingCloseParen
from .errors import MissingCloseBrack, MissingOps, InvalidPc, MissingPc
from .errors import TooBigForSingle, TooBigForDouble
from .tokens import Address, Register, IntegerTok, Symbol, Instruction
from .tokens import RegAddress, Label, NewSymbol, Section, DataType
from .tokens import StringTok, Comma, DupTok, QuestionTok
from .tokens import OpenParen, CloseParen, OpenBracket, CloseBracket
from .tokens import PlusTok, MinusTok, ConstantSign
from .tokens import FloatTok
from .virtual_machine import MEM_SIZE

TOKENS = 0
CODE = 1
LINE_NUMBER = 2

DONT_INIT = "?"

MAX_BYTE = 255
MAX_SHORT = 65535
MAX_LONG = 4294967295

BYTES = 0
MAX_VAL = 1
dtype_info = {
    "DB": (1, MAX_BYTE),
    "DW": (MEM_SIZE / 16, MAX_SHORT),   # we should revisit this choice
    "DD": (MEM_SIZE / 8, MAX_LONG),
    "REAL4": (1, MAX_BYTE),
    "REAL8": (1, MAX_BYTE),
    "DBL": (1, MAX_BYTE),
    "FL": (1, MAX_BYTE)
}

PARAM_TYPE = 1
PARAM_VAL = 0


def add_debug(s, vm):
    vm.debug += (s + "\n")


def minus_token(token_line, pos, line_num):
    """
    Negates the next token if minus token found

    Args:
        token_line: Line of code
        pos: Position of minus token
    """
    try:
        if isinstance(token_line[pos + 1], IntegerTok):
            token_line[pos + 1].negate_val()
        else:
            raise InvalidArgument("-", line_num)
    except Exception:
        raise InvalidArgument("-", line_num)


def register_token(token_line, pos, vm, line_num):
    if vm.flavor == "intel" or vm.flavor == "att":

        return (token_line[pos], pos + 1)
    elif (pos + 1 < len(token_line) and
          isinstance(token_line[pos + 1], OpenParen)):
        reg = None
        disp = None
        reg, disp, pos = get_address_mips(token_line, pos + 2, vm, line_num,
                                          token_line[pos])
        return (RegAddress(reg.get_nm(), vm, disp), pos)
    else:
        return (token_line[pos], pos + 1)


def number_token(token_line, pos, vm, line_num):
    """
    If token seen is an integer, determine by flavor whether
    to return the integer token or return an address token

    Args:
        token_line: Line of code
        pos: Position where integer token is seen
        vm: Virtual machine

    Returns:
        Integer or address token, next positon to look at
    """
    if vm.flavor == "intel":

        return (token_line[pos], pos + 1)
    elif (pos + 1 < len(token_line) and
          isinstance(token_line[pos + 1], OpenParen)):
        reg = None
        disp = None
        if vm.flavor == "att":
            reg, disp, pos = get_address_att(token_line, pos + 2,
                                             vm, line_num,
                                             token_line[pos].get_val(line_num))
        else:
            reg, disp, pos = get_address_mips(
                token_line, pos + 2,
                vm, line_num,
                token_line[pos].get_val(line_num))
        if reg:
            return (RegAddress(reg.get_nm(), vm, disp,
                               reg.get_multiplier()), pos)
        else:
            return (Address(hex(disp).split('x')[-1].upper(), vm), pos)
    elif vm.flavor == "att" and token_line[pos].con is False:
        mem_val = token_line[pos].get_val(line_num)
        return (Address(hex(mem_val).split('x')[-1].upper(), vm), pos + 1)
    else:
        return (token_line[pos], pos + 1)


def symbol_token(token_line, pos, vm, line_num):
    """
    If token seen is a symbol, determine by flavor whether
    to return the symbol token or return an address token

    Args:
        token_line: Line of code
        pos: Position where integer token is seen
        vm: Virtual machine

    Returns:
        Symbol or address token, next positon to look at
    """
    if vm.flavor != "mips_asm":
        return (Symbol(token_line[pos].get_nm(), vm), pos + 1)
    elif (pos + 1 < len(token_line) and
          isinstance(token_line[pos + 1], OpenParen)):
        reg, disp, pos = get_address_mips(token_line, pos + 2, vm, line_num,
                                          vm.symbols[token_line[pos].get_nm()])
        return (RegAddress(reg.get_nm(), vm, disp,
                           reg.get_multiplier()), pos)
    else:
        return (Symbol(token_line[pos].get_nm(), vm), pos + 1)


def is_start_address(token_line, pos, flavor):
    """
    Determines at current location if there is an address

    Args:
        token_line: Line of code
        pos: Position of possible address
        flavor: Assembly langauge

    Returns:
        True if address seen, False otherwise
    """
    if isinstance(token_line[pos], OpenParen) and flavor == "att":
        return True
    elif (isinstance(token_line[pos], OpenParen) and
          (flavor == "mips_asm" or flavor == "mips_mml")):
        return True
    elif isinstance(token_line[pos], OpenBracket) and flavor == "intel":
        return True
    else:
        return False


def get_data_type(token_line, pos, line_num):
    """
    Returns the data type

    Args:
        token_line: List of data tokens
        pos: Position of list

    Returns:
        Data type of variable
    """
    if pos >= len(token_line):
        raise MissingData(line_num)
    elif isinstance(token_line[pos], DataType):
        if token_line[pos].get_nm() not in dtype_info:
            raise InvalidDataType(token_line[pos].get_nm(), line_num)
        else:
            return token_line[pos].get_nm()
    else:
        raise InvalidDataType(token_line[pos].get_nm(), line_num)


def get_data_token(token_line, pos, line_num):
    """
    Returns the data token found

    Args:
        token_line: Line of code
        pos: Position to find data token

    Returns:
        Data value, next position
    """
    if pos >= len(token_line):
        raise MissingData(line_num)
    if isinstance(token_line[pos], MinusTok):
        try:
            if isinstance(token_line[pos + 1], IntegerTok):
                token_line[pos + 1].negate_val()
                return get_data_token(token_line, pos + 1, line_num)
            else:
                raise InvalidDataVal("-", line_num)
        except Exception:
            raise InvalidDataVal("-", line_num)
    elif isinstance(token_line[pos], IntegerTok):
        return token_line[pos].get_val(line_num), pos + 1
    elif isinstance(token_line[pos], FloatTok):
        if (token_line[pos].get_type() == ".float" and
                token_line[pos].get_val(line_num) > float(2 ** 22)):
            raise TooBigForSingle(str(token_line[pos].get_val(line_num)),
                                  line_num)
        if (token_line[pos].get_type() == ".double" and
                token_line[pos].get_val(line_num) > float(2 * 10 ** 14)):
            raise TooBigForDouble(str(token_line[pos].get_val(line_num)),
                                  line_num)
        return token_line[pos].get_val(line_num), pos + 1
    elif isinstance(token_line[pos], QuestionTok):
        return DONT_INIT, pos + 1
    else:
        raise InvalidArgument(token_line[pos].get_nm(), line_num)


def get_DUP_value(token_line, pos, line_num):
    """
    Finds the value to duplicate if found

    Args:
        token_line: List of data tokens
        pos: Beginning position to parse for DUP

    Returns:
        Integer value to duplicate, next position
    """
    if isinstance(token_line[pos], OpenParen):
        pos += 1
    else:
        raise MissingOpenParen(line_num)
    dup_value, pos = get_data_token(token_line, pos, line_num)
    if isinstance(token_line[pos], CloseParen):
        return dup_value, pos + 1
    else:
        raise MissingCloseParen(line_num)


def is_str_termin(token_line, pos, line_num):
    if not isinstance(token_line[pos], IntegerTok):
        return False
    # check if following is an Integer with value of 0
    elif (token_line[pos].get_val(line_num) != 0):
        return False
    return True


NUM_STR_TOKS = 3
COMMA_POS = 1
TERM_POS = 2


def parse_string_token(token_line, pos, line_num):
    """
    Creates a list of the string token's ASCII values

    Args:
        token_line: List of data tokens
        pos: Position of string token

    Returns:
        List of ASCII values, followed by a 0
    """
    if len(token_line) < pos + NUM_STR_TOKS:
        raise MissingData(line_num)
    elif not isinstance(token_line[pos + COMMA_POS], Comma):
        raise MissingComma(line_num)
# is string terminated properly?
    elif not is_str_termin(token_line, pos + TERM_POS, line_num):
        raise InvalidDataVal(str(token_line[pos + TERM_POS].get_val(line_num)),
                             line_num)

    else:
        ascii_list = []
        for letter in token_line[pos].get_nm():
            if letter != "'":
                ascii_list.append(ord(letter))
        ascii_list.append(0)
        return (ascii_list, pos + 3)


def parse_dup_token(token_line, data_type, pos, line_num):
    """
    Parses integer token followed by DUP token

    Args:
        token_line: List of data tokens
        data_type: Data type of variable
        pos = Position of integer token

    Returns:
        List of duplicated values
    """
    dup_list = None
    duplicate = token_line[pos].get_val(line_num)
    value, pos = get_DUP_value(token_line, pos + 2, line_num)
    if value == DONT_INIT:
        dup_list = [randrange(0, dtype_info[data_type][MAX_VAL])] * duplicate
    else:
        dup_list = [value] * duplicate
    return (dup_list, pos)


def get_values(token_line, data_type, pos, values_list, line_num):
    """
    Creates a list of values for each variable when it is declared.

    Args:
        token_line: List of data tokens
        data_type: Data type of variable
        pos: Beginning pos to parse from
        values_list: List of values

    Returns:
        List of integer values
    """
    if pos >= len(token_line):
        raise MissingData(line_num)
    elif isinstance(token_line[pos], StringTok):
        ascii_list, pos = parse_string_token(token_line, pos, line_num)
        values_list.extend(ascii_list)
    else:
        first_data, pos = get_data_token(token_line, pos, line_num)
        if first_data == DONT_INIT:
            values_list.append(randrange(0, dtype_info[data_type][MAX_VAL]))
        else:
            values_list.append(first_data)
    if pos >= len(token_line):
        return values_list, pos
    else:
        next_term = token_line[pos]
        if isinstance(next_term, DupTok):
            values_list.pop()
            dup_list, pos = parse_dup_token(token_line, data_type,
                                            pos - 1, line_num)
            values_list.extend(dup_list)
            try:
                next_term = token_line[pos]
            except Exception:
                return values_list, pos
        if isinstance(next_term, Comma):
            return get_values(token_line, data_type, pos + 1,
                              values_list, line_num)
        else:
            raise InvalidDataVal(token_line[pos].get_nm(), line_num)


def parse_data_token(token_line, vm, mem_loc, line_num):
    """
    Parses data tokens, assigns each value to a memory location

    Args:
        token_line: List of data tokens
        vm: Virtual machine
        mem_loc: Starting memory storage location

    Returns:
        Returns the next memory location to be used
    """
    pos = 0
    symbol = ""
    data_vals = []
    if not isinstance(token_line[pos], NewSymbol):
        raise InvalidArgument(token_line[pos].get_nm(), line_num)
    else:
        symbol = token_line[pos].get_nm()
    data_type = get_data_type(token_line, pos + 1, line_num)
    pos += 2
    data_vals, pos = get_values(token_line, data_type, pos,
                                data_vals, line_num)

    # store memory location
    vm.symbols[symbol] = mem_loc
    add_debug("Symbol table now holds " + str(mem_loc), vm)
    for value in data_vals:
        if vm.get_data_init() == "on":
            vm.memory[hex(mem_loc).split('x')[-1].upper()] = value
        if vm.flavor == "mips_asm" or vm.flavor == "riscv":
            mem_loc += 4
        else:
            mem_loc += 1
    return mem_loc


def get_term(token_line, pos, vm, line_num):
    """
    Returns the next term of the expression

    Args:
        token_line: Line of code
        pos: Position of next term to be found
        vm: Virtual machine

    Returns:
        Next term token, position of token
    """
    # call function again to get the negated term
    if isinstance(token_line[pos], MinusTok):
        try:
            token_line[pos + 1].negate_val()
            return get_term(token_line, pos + 1, vm, line_num)
        except Exception:
            raise InvalidArgument(token_line[pos].get_nm(), line_num)
    # integer or register term
    elif (isinstance(token_line[pos], IntegerTok) or
          isinstance(token_line[pos], Register)):
        return (token_line[pos], pos)
    # symbol term
    elif isinstance(token_line[pos], NewSymbol):
        if token_line[pos].get_nm() in vm.symbols:
            return (Symbol(token_line[pos].get_nm(), vm), pos)
        else:
            raise InvalidMemLoc(token_line[pos].get_nm(), line_num)
    else:
        raise InvalidMemLoc(token_line[pos].get_nm(), line_num)


REG = 0
DISP_VAL = 1
POSITION = 2


def get_expr_intel(token_line, pos, vm, line_num, reg):
    """
    Returns the register and the evaluated expression

    Args:
        token_line: Line of code
        pos: Position of address expression
        vm: Virtual machine

    Returns:
        Register token, displacement, next position
    """
    if len(token_line) < pos + 2:
        return MissingOps(line_num)
    left, pos = get_term(token_line, pos, vm, line_num)
    if isinstance(left, Register):
        reg = left
    next_term = token_line[pos + 1]
    if isinstance(next_term, PlusTok):
        next_val_pos = get_expr_intel(token_line, pos + 2, vm, line_num, reg)
        return (next_val_pos[REG],
                left.get_val(line_num) + next_val_pos[DISP_VAL],
                next_val_pos[POSITION])
    elif isinstance(next_term, MinusTok):
        next_val_pos = get_expr_intel(token_line, pos + 2, vm, line_num, reg)
        return (next_val_pos[REG],
                left.get_val(line_num) - next_val_pos[DISP_VAL],
                next_val_pos[POSITION])
    else:
        return (reg, left.get_val(line_num), pos + 1)


SEC_REG = 0


def get_expr_att(token_line, pos, vm, line_num, reg, disp_list):
    """
    Returns address expression for AT&T

    Args:
        token_line: Line of code
        pos: Position of address
        vm: Virtual machine
        disp_list: List of displacements

    Returns:
        Register token, displacement(s), next position
    """
    if len(token_line) < pos + 2:
        return MissingOps(line_num)
    left, pos = get_term(token_line, pos, vm, line_num)

    # Retrieved Register Term
    if isinstance(left, Register) and reg is None:
        reg = left
    elif isinstance(left, Register) and disp_list[SEC_REG] is None:
        disp_list[SEC_REG] = left

    # at most two registers allowed
    elif isinstance(left, Register):
        raise InvalidMemLoc(token_line[pos].get_nm(), line_num)

    # Retrieved Integer term
    elif isinstance(left, IntegerTok):
        if token_line[pos - 2] == reg:
            reg.set_multiplier(left.get_val(line_num))
        elif token_line[pos - 2] == disp_list[SEC_REG]:
            disp_list[SEC_REG].set_multiplier(left.get_val(line_num))
        else:
            disp_list.append(left.get_val(line_num))

    # Retrieved Symbol term
    else:
        disp_list.append(vm.symbols[token_line[pos].get_nm()])
    next_term = token_line[pos + 1]
    if isinstance(next_term, Comma):
        return get_expr_att(token_line, pos + 2, vm, line_num, reg, disp_list)
    else:
        return (reg, disp_list, pos + 1)


def get_expr_mips(token_line, pos, vm, line_num):
    """
    Returns address expression for MIPS

    Args:
        token_line: Line of code
        pos: Position of address
        vm: Virtual machine

    Returns:
        Token term, next position
    """

    if pos >= len(token_line):
        raise MissingOps(line_num)
    left, pos = get_term(token_line, pos, vm, line_num)
    if isinstance(left, Register):
        return (left, pos + 1)
    else:
        raise InvalidMemLoc(left.get_nm(), line_num)


def get_address_intel(token_line, pos, vm, line_num):
    """
    Converts a sublist of the tokenized instruction into
    corresponding address token

    Args:
        token_line: List of instruction tokens
        pos: Beginning position in list
        vm: Virtual machine

    Returns:
        Register token, displacement, next position
    """

    if pos >= len(token_line):
        raise InvalidMemLoc("", line_num)
    reg = None
    disp = 0
    reg, disp, pos = get_expr_intel(token_line, pos, vm, line_num, reg)
    if pos >= len(token_line):
        raise MissingCloseBrack(line_num)
    elif isinstance(token_line[pos], CloseBracket):
        return (reg, disp, pos + 1)
    else:
        raise InvalidMemLoc(token_line[pos].get_nm(), line_num)


def get_address_att(token_line, pos, vm, line_num, disp=0):
    """
    Converts a sublist of the tokenized instruction into
    corresponding address token for AT&T

    Args:
        token_line: List of instruction tokens
        pos: Beginning position in list
        vm: Virtual machine
        Disp: Numeric displacement

    Returns:
        Register token, displacement(s), next position
    """
    if pos >= len(token_line):
        raise InvalidMemLoc("", line_num)
    reg = None
    reg, disp_list, pos = get_expr_att(token_line, pos, vm,
                                       line_num, reg, [None])
    if pos >= len(token_line):
        raise MissingCloseParen(line_num)
    elif isinstance(token_line[pos], CloseParen):
        if len(disp_list) > 1:
            for values in range(1, len(disp_list)):
                disp += disp_list[values]
        if disp_list[0] is None:
            return (reg, disp, pos + 1)
        return (reg, [disp_list[0], disp], pos + 1)
    else:
        raise InvalidMemLoc(token_line[pos].get_nm(), line_num)


def get_address_mips(token_line, pos, vm, line_num, disp=0):
    """
    Converts a sublist of the tokenized instruction into
    corresponding address token for MIPS

    Args:
        token_line: List of instruction tokens
        pos: Beginning position in list
        vm: Virtual machine

    Returns:
        Register token, displacement, next position
    """

    if pos >= len(token_line):
        raise InvalidMemLoc("", line_num)
    reg, pos = get_expr_mips(token_line, pos, vm, line_num)
    if pos >= len(token_line):
        raise MissingCloseParen(line_num)
    elif isinstance(token_line[pos], CloseParen):
        return (reg, disp, pos + 1)
    else:
        raise InvalidMemLoc(token_line[pos].get_nm(), line_num)


def get_address_location(token_line, pos, vm, line_num):
    """
    Retrieves address at current position in code
    Retrieves address by coding language

    Args:
        token_line: List of the tokenized instruction
        pos: Beginning pos of list
        vm: Virtual machine

    Returns:
        RegAddress or Address token,
        position of next item in instruction
    """
    reg = None
    disp = 0
    if vm.flavor == "intel":
        reg, disp, pos = get_address_intel(token_line, pos, vm, line_num)
    elif vm.flavor == "att":
        reg, disp, pos = get_address_att(token_line, pos, vm, line_num)
    else:
        reg, disp, pos = get_address_mips(token_line, pos, vm, line_num)
    if reg:
        return (RegAddress(reg.get_nm(), vm,
                           disp, reg.get_multiplier()), pos)
    else:
        # eliminates negative memory locations
        if disp < 0:
            raise InvalidMemLoc(str(disp), line_num)
        return (Address(hex(disp).split('x')[-1].upper(), vm), pos)


def check_constant(token_line, pos, line_num):
    """
    Checks if the following token after constant token is valid

    Args:
        token_line: Line of code
        pos: Position of constant token

    Returns:
        True if valid, False otherwise
    """
    try:
        if (not isinstance(token_line[pos + 1], MinusTok) and
                not isinstance(token_line[pos + 1], IntegerTok)):
            return False
        if isinstance(token_line[pos + 1], IntegerTok):
            token_line[pos + 1].con = True
        else:
            try:
                token_line[pos + 2].con = True
            except Exception:
                raise InvalidArgument("-", line_num)
        return True
    except Exception:
        raise InvalidArgument("$", line_num)


def get_op(token_line, pos, vm, line_num):
    """
    Retrieves operand of instruction

    Args:
        token_line: List of the tokenized instruction
        pos: Beginning pos of list
        vm: Virtual machine

    Returns:
        Operand token, position of next item in instruction
    """
    if pos >= len(token_line):
        raise MissingOps(line_num)

# Register
    elif isinstance(token_line[pos], Register):
        return register_token(token_line, pos, vm, line_num)

# Floating Point Token
    elif isinstance(token_line[pos], FloatTok):
        return token_line[pos], pos+1

# Constant Token
    elif isinstance(token_line[pos], ConstantSign):
        if vm.flavor == "att" and check_constant(token_line, pos, line_num):
            return get_op(token_line, pos + 1, vm, line_num)
        else:
            raise InvalidArgument("$", line_num)

# Minus Token
    elif isinstance(token_line[pos], MinusTok):
        minus_token(token_line, pos, line_num)
        return get_op(token_line, pos + 1, vm, line_num)

# Integer Token
    elif isinstance(token_line[pos], IntegerTok):
        return number_token(token_line, pos, vm, line_num)

# Symbol/Label Token
    elif isinstance(token_line[pos], NewSymbol):
        if vm.flavor == "wasm":
            return token_line[pos], pos + 1
        else:
            if token_line[pos].get_nm() in vm.labels:
                return (Label(token_line[pos].get_nm(), vm), pos + 1)
            elif token_line[pos].get_nm() in vm.symbols:
                return symbol_token(token_line, pos, vm, line_num)
            elif vm.flavor == 'intel' and token_line[pos].get_nm()[:2] == "ST":
                return token_line[pos], pos + 1
            else:
                raise UnknownName(token_line[pos].get_nm(), line_num)

# Address Token
    elif is_start_address(token_line, pos, vm.flavor):
        return get_address_location(token_line, pos + 1, vm, line_num)
    else:
        raise InvalidArgument(token_line[pos].get_nm(), line_num)


def get_op_list(token_line, pos, vm, op_lst, line_num):
    """
    Returns a list of ops

    Args:
        token_line: Line of code
        pos: Starting position to retrieve op
        vm: Virtual machine
        op_lst: List of ops

    Returns:
        A list of ops, next position
    """
    op, pos = get_op(token_line, pos, vm, line_num)
    op_lst.append(op)
    if pos >= len(token_line):
        return op_lst, pos
    else:
        next_op = token_line[pos]
        if isinstance(next_op, Comma):
            return get_op_list(token_line, pos + 1, vm, op_lst, line_num)
        else:
            raise MissingComma(line_num)


def get_pc(token_line, pos, line_num):
    """
    Returns the PC counter of the instruction

    Args:
        token_line: Line of code
        pos: Position of PC counter

    Returns:
        Integer token of PC counter value
    """
    if not isinstance(token_line[pos], IntegerTok):
        raise MissingPc(line_num)
    elif token_line[pos].get_val(line_num) % 4 != 0:
        raise InvalidPc(str(token_line[pos].get_val(line_num)), line_num)
    else:
        return token_line[pos]


def parse_exec_unit(token_line, vm, line_num):
    """
    Parses instruction

    Args:
        token_line: Tokenized instruction
        vm: Virtual machine

    Returns:
        List of tokens: instruction, operand(s)
        If MIPS: PC, instruction, operand(s)
    """
    pos = 0
    token_instruction = []
    op_lst = []
    # retrieve PC counter
    if (vm.flavor == "mips_asm" or vm.flavor == "mips_mml" or
            vm.flavor == "riscv"):
        token_instruction.append(get_pc(token_line, pos, line_num))
        pos += 1

    # retrieve instruction
    if not isinstance(token_line[pos], Instruction):
        raise InvalidInstruction(token_line[pos].get_nm(), line_num)
    token_instruction.append(token_line[pos])
    pos += 1

    # retrieve ops
    if pos < len(token_line):
        op_lst, pos = get_op_list(token_line, pos, vm, op_lst, line_num)
    token_instruction.extend(op_lst)
    # switch ops if flavor is AT&T
    if vm.flavor == 'att' and len(token_instruction) > 2:
        switch_vals = token_instruction[1], token_instruction[2]
        token_instruction[1] = switch_vals[1]
        token_instruction[2] = switch_vals[0]

    return token_instruction


def parse(tok_lines, vm, web):
    """
    Parses the analysis obtained from lexical analysis

    Args:
        tok_lines: Lines containing each line of code
        vm: Virtual machine
        web: Boolean indicating whether source is from the website
             or from kernel

    Returns:
        A list of parsed instructions
    """
    parse_data = False
    parse_text = True
    token_instrs = []
    mem_loc = 0
    ip_init = None
    for line in tok_lines:
        tokens = line[TOKENS]
        line_num = line[LINE_NUMBER]
        if isinstance(tokens[0], Section):
            if tokens[0].get_nm() == "data":
                parse_data = True
                parse_text = False
                if not web:
                    vm.set_data_init("on")
                    vm.re_init()
                continue
            elif tokens[0].get_nm() == "text":
                parse_text = True
                parse_data = False
                continue
            else:
                raise InvalidSection(tokens[0].get_nm(), line_num)
        if parse_data:
            mem_loc = parse_data_token(tokens, vm, mem_loc, line_num)
        elif parse_text:
            vm.set_data_init("off")
            parsed_unit = parse_exec_unit(tokens, vm, line_num)
            token_instrs.append((parsed_unit, line[CODE], line_num))
            if (vm.flavor == "mips_asm" or
                vm.flavor == "mips_mml" or
                    vm.flavor == "riscv") and ip_init is None:
                ip_init = token_instrs[0][TOKENS][0].get_val(line_num)
                vm.start_ip = ip_init
    return token_instrs



from __future__ import print_function
# from assembler.virtual_machine import intel_machine
from assembler.errors import DivisionZero
from collections import namedtuple
from fractions import Fraction
import struct

DEFAULT_SIZE = (11, 52)


def trunc_round(n, k):
    """
    :param n: int
    :param k: int
    :return: int after
    trunc_round(49,2) -> 3
    """
    rshift = n.bit_length() - 1 - k
    if rshift >= 0:
        n >>= (rshift)
    else:
        n <<= (-rshift)
    return (n + 1) >> 1


def more_bin_digits(n, k):
    """

    :param n: int
    :param k: int
    :return: boolean
    """
    return bool(n >> k)


def unset_high_bit(n):
    assert n > 0
    return n ^ (1 << (n.bit_length() - 1))


def fbin(n, nbits):
    assert (0 <= n)
    assert not (n >> nbits)
    return "{val:0>{width}}".format(val=bin(n)[2:], width=nbits)


_anyfloat = namedtuple("anyfloat", "sign exponent significand")


class anyfloat(_anyfloat):

    _b32 = 1 << 32
    _b64 = 1 << 64

    def __new__(cls, sign, exponent, significand):
        assert sign in (0, 1)
        if significand:
            significand = significand//(significand & -significand)
        return _anyfloat.__new__(cls, sign, exponent, significand)

    @staticmethod
    def _encode(log2, mantissa, a, b):
        A = ~(~0 << a)
        AA = A >> 1
        if mantissa <= 0:
            encoded = (A, 0) if (mantissa == -1) else (A, 1 << (b-1))
            return encoded if mantissa else (0, 0)
        elif log2 <= - AA:
            nbits = b + log2 + AA
            rounded = trunc_round(mantissa, nbits) if (nbits >= 0) else 0
            return (1, 0) if more_bin_digits(rounded, b) else (0, rounded)
        elif log2 <= AA:
            rounded = trunc_round(mantissa, b + 1)
            encoded = ((log2 + 1 + AA, 0) if (log2 < AA) else (A, 0))
            if more_bin_digits(rounded, b+1):
                return encoded
            else:
                return (log2 + AA, unset_high_bit(rounded))
        else:
            return (A, 0)

    @staticmethod
    def _decode(exponent, significand, a, b):
        A = ~(~0 << a)
        AA = A >> 1
        assert 0 <= exponent <= A
        assert 0 <= significand < (1 << b)
        if exponent == A:
            return (0, -2 if significand else -1)
        elif exponent:  # normal case
            return (exponent - AA, significand | (1 << b))
        else:  # subnormal case
            if significand:
                return (significand.bit_length() - AA - b, significand)
            else:
                return (0, 0)

    def __float__(self):
        return self.int64_to_float(self.to_ieee())

    @classmethod
    def from_float(cls, x):
        """Create an anyfloat instance from a python float
        (64 bits double precision number)."""
        val = cls.from_ieee(cls.float_to_int64(x))
        return val

    @classmethod
    def from_ieee(cls, n, size=DEFAULT_SIZE):
        """Create an anyfloat from an ieee754 integer.

        Create an anyfloat from an integer which binary representation
        is the ieee754 format of a floating point number.
        The argument 'size' is a tuple (w, p)
        containing the width of the exponent part and the significand part in
        this ieee754 format."""
        w, p = size
        r = n >> p
        significand = (r << p) ^ n
        sign = int(r >> w)
        if sign not in (0, 1):
            error_msg = "Integer value out of range for ieee754 format"
            raise ValueError((error_msg, n, size))
        exponent = (sign << w) ^ r
        e, s = cls._decode(exponent, significand, w, p)
        if e == -2:
            sign = 0
        cls.sign = sign
        cls.exponent = e
        cls.significand = s
        return cls(sign, e, s)

    def ieee_parts(self, size=DEFAULT_SIZE):
        w, p = size
        e, s = self._encode(self.exponent, self.significand, w, p)
        sign = 0 if (e + 1) >> w else self.sign
        return [sign, e, s]

    def abs_sign(self):
        self.sign = 0  # changes sign to  0 (positive)

    def change_sign(self):
        if self.sign == 0:
            self.sign = 1
        else:
            self.sign = 0

    def to_ieee(self, size=DEFAULT_SIZE):
        """Convert to an ieee754 integer.

        Convert self to an integer which binary representation
        is the ieee754 format corresponding
        to the 'size' argument (read the documentation of from_ieee()
        for the meaning of the size argument.
        """
        sign, e, s = self.ieee_parts(size)
        return (((sign << size[0]) | e) << size[1]) | s

    @classmethod
    def int64_to_float(cls, n):
        """Convert a 64 bits integer to a python float.

        This class method converts an integer
        representing a 64 bits floating point
        number in the ieee754 double precision format
        to this floating point number."""

        if not (0 <= n < cls._b64):
            error_msg = "Integer value out of range for 64 bits ieee754 format"
            raise ValueError((error_msg, n))
        u, v = divmod(n, cls._b32)
        return struct.unpack(">d", struct.pack(">LL", u, v))[0]

    @classmethod
    def float_to_int64(cls, x):
        """Convert a python float to a 64 bits integer.

        This class method converts a float to an integer representing this
        float in the 64 bits ieee754 double precision format."""

        u, v = struct.unpack(">LL", struct.pack(">d", x))

        return (u << 32) | v

    def bin(self, size=DEFAULT_SIZE, sep=' '):
        """Return a binary representation of self.

        The returned string contains only the characters '0' and '1'
        and shows the ieee754 representation of the real number
        corresponding to self within the given size = (w, p).
        """
        if sep:
            sign, e, s = self.ieee_parts(size)
            bin_left = fbin(sign, 1)
            bin_mid = fbin(e, size[0])
            bin_right = fbin(s, size[1])
            return sep.join((bin_left, bin_mid, bin_right))
        else:
            return fbin(self.to_ieee(size), sum(size) + 1)

    def to_fraction(self):
        s = self.significand
        b = s.bit_length()
        k = self.exponent + 1 - b
        if k < 0:
            d = Fraction(s, 2 ** (-k))
        else:
            d = Fraction(s * 2**k, 1)
        return -d if self.sign else d


def binaryAdd(bin1, bin2):
    additionBin = ""
    carryOver = 0
    i = -1
    while i != (len(bin1) * -1) - 1:
        if bin1[i] == '.':
            additionBin = '.' + additionBin
            i -= 1
            continue
        sum = int(bin1[i]) + int(bin2[i]) + carryOver
        if sum == 1 or sum == 0:
            additionBin = str(sum) + additionBin
            carryOver = 0
        elif sum == 2:
            additionBin = str(0) + additionBin
            carryOver = 1
        else:
            additionBin = str(1) + additionBin
            carryOver = 1
        i -= 1
    if carryOver != 0:
        additionBin = str(carryOver) + additionBin
    return additionBin


def binarySubtract(val1, val2):
    """

    :param val1: binary representation of int value in String format
    :param val2: binary representation of int value in String format
    :return: binary value in string format


    """
    difference = [0 for x in val1 if x != '.']
    bin1 = [int(x) for x in val1 if x != '.']
    bin2 = [int(x) for x in val2 if x != '.']
    for i in range(1, len(bin1) + 1):
        value = (bin1[-i]) - (bin2[-i])
        if value == -1:
            j = -i
            while bin1[j] == 0:
                j -= 1
            bin1[j] -= 1
            for k in range(j + 1, -i):
                bin1[k] = 1
            bin1[-i] = 2
            value = (bin1[-i]) - (bin2[-i])
        difference[-i] = value
    difference.insert(1, '.')
    binary = ''.join(str(x) for x in difference)
    return binary


def convertFraction(mantissa):
    product = 0
    for i in range(len(mantissa)):
        product += (int(mantissa[i])*(2**(-i-1)))
    return product


def convertFromIEE(IEEE):
    bits = IEEE.split(" ")
    sign, expo, mantissa = bits[0], bits[1], bits[2]

    fraction = convertFraction(mantissa)
    float = (1 + fraction) * (2**(int(expo, 2)))
    if sign == '0':
        return float
    return -1 * float


def fabs(val):
    IEEE = anyfloat.from_float(val)
    IEEE.abs_sign()
    return float(IEEE)


def chs(val):
    IEEE = anyfloat.from_float(val)
    IEEE.change_sign()
    return float(IEEE)


# performing addition of 2  numbers
def add(val1, val2):
    val1, val2 = float(val1), float(val2)

    if abs(val2) > abs(val1):
        val1, val2 = val2, val1
    X1 = anyfloat.from_float(val1)
    sign1, expo1 = X1.sign, X1.exponent
    mantissaBin1 = X1.bin(size=(11, 52)).split(" ")[2]
    X2 = anyfloat.from_float(val2)
    sign2, expo2 = X2.sign, X2.exponent
    mantissaBin2 = X2.bin(size=(11, 52)).split(" ")[2]
    expoDiff = expo1 - expo2
    if expoDiff != 0:
        allignedMant2 = "1"+mantissaBin2
        for i in range(abs(expoDiff)-1):
            allignedMant2 = "0"+allignedMant2
            allignedMant2 = allignedMant2[:-1]
        allignedMant2 = "0."+allignedMant2
        allignedMant2 = allignedMant2[:-1]
    else:
        allignedMant2 = "1."+mantissaBin2
    allignedMant1 = '1.'+mantissaBin1
    if sign1 == sign2:  # add
        allignedMant3 = binaryAdd(allignedMant1, allignedMant2)
    else:
        allignedMant3 = binarySubtract(allignedMant1, allignedMant2)
    decimalPlace = allignedMant3.find(".")
    if decimalPlace != 1:
        expo1 += (decimalPlace - 1)
        rightSide = allignedMant3[1:decimalPlace]
        allignedMant3_left = allignedMant3[0] + "." + rightSide
        allignedMant3_right = allignedMant3[decimalPlace + 1:-len(rightSide)]
        allignedMant3 = allignedMant3_left + allignedMant3_right
    if allignedMant3[0] == '0':
        firstOne = allignedMant3.find("1")
        expo1 -= (firstOne - 1)
        allignedMant3 = allignedMant3[firstOne]+'.'+allignedMant3[firstOne+1:]
        while len(allignedMant3) < 54:
            allignedMant3 += '0'
    IEEE = str(sign1) + " " + bin(expo1) + " " + allignedMant3[2:]
    return convertFromIEE(IEEE)


def sub(val1, val2):  # subtraction is the same as addition, (x-y) = (x+(-y))
    val2 = val2 * -1
    return add(val1, val2)


# performing  multiplication of two binary numbers
def binaryMultiply(bin1, bin2):
    bin1 = bin1.strip(".")
    bin2 = bin2.strip(".")
    binList1 = [int(x) for x in bin1]
    binList2 = [int(x) for x in bin2]
    product = []
    for i in range(1, len(binList2) + 1):
        if binList2[-i] == 1:
            arr = [x for x in binList1]
            extraZero = [0] * (i - 1)
            arr.extend(extraZero)
            product.append(arr)
    maxLen = len(product[-1])
    for i in range(len(product)):
        while len(product[i]) != maxLen:
            product[i] = [0] + product[i]
    for i in range(len(product)):
        product[i] = "".join(str(e) for e in product[i])
    while len(product) > 1:
        maxLen = max([len(x) for x in product])
        val1 = product[0]
        val2 = product[1]
        while len(val1) < maxLen:
            val1 = '0' + val1
        while len(val2) < maxLen:
            val2 = '0' + val2
        val3 = binaryAdd(val1, val2)
        product.append(val3)
        product = product[2:]
    return product[0]


def mul(val1, val2):
    if abs(val2) > abs(val1):
        val1, val2 = val2, val1
    if val1 == 0 or val2 == 0:
        return 0
    multiplier = anyfloat.from_float(val1)
    multiplierSign, multiplierExpo = multiplier.sign, multiplier.exponent
    multiplierMantissaBin = multiplier.bin(size=(11, 52)).split(" ")[2]
    multiplicand = anyfloat.from_float(val2)
    multiplicandSign = multiplicand.sign
    multiplicandExpo = multiplicand.exponent
    multiplicandMantissaBin = multiplicand.bin(size=(11, 52)).split(" ")[2]
    productSign = multiplierSign ^ multiplicandSign  # xor
    productExponent = multiplierExpo + multiplicandExpo
    allignMultiplierMant = "1" + multiplierMantissaBin
    allignMultiplicMant = '1' + multiplicandMantissaBin
    productMantissa = binaryMultiply(allignMultiplierMant, allignMultiplicMant)
    productMantissa = productMantissa[0]+'.'+productMantissa[1:-1]
    productExponent += (len(productMantissa) - 105)
    IEEE = str(productSign)+" "+bin(productExponent)+" "+productMantissa[2:]
    returnVal = convertFromIEE(IEEE)
    return returnVal


def div(val1, val2):
    """There is a faster way to do division.  Its called
   division by reciprocal approximation.  It takes about the same
   time as a fl. pt. multiply.  """
    if val2 == 0.0:
        raise DivisionZero()
    else:
        return mul(val1, 1 / val2)



from jupyter_client.kernelspec import KernelSpecManager

kernel_name = "riscv"


def uninstall_kernel():
    print("Uninstalling", kernel_name, "kernel...")
    KernelSpecManager().remove_kernel_spec(kernel_name)
    print(kernel_name.capitalize(), "kernel uninstallation complete")


if __name__ == '__main__':
    uninstall_kernel()



"""
arithmetic.py: arithmetic and logic instructions.
"""

import operator as opfunc

from assembler.errors import DivisionZero, check_num_args, InvalidConVal
from assembler.errors import InvalidOperand
from assembler.tokens import Instruction, MAX_INT, Register
from assembler.ops_check import one_op_arith


def two_op_arith(ops, vm, instr, line_num, operator):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 2, line_num)
    ops[0].set_val(
        checkflag(operator(ops[0].get_val(line_num),
                           ops[1].get_val(line_num)), vm), line_num)
    vm.changes.add(ops[0].get_nm())


def checkflag(val, vm):
    if(val > MAX_INT):
        vm.flags['CF'] = 1
        val = val - MAX_INT+1
    else:
        vm.flags['CF'] = 0
    return val


class Add(Instruction):
    """
        <instr>
             add
        </instr>
        <syntax>
            ADD reg, reg
            ADD reg, mem
            ADD reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.add)


class Sub(Instruction):
    """
        <instr>
             sub
        </instr>
        <syntax>
            SUB reg, reg
            SUB reg, mem
            SUB reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.sub)


class Imul(Instruction):
    """
        <instr>
             imul
        </instr>
        <syntax>
            IMUL reg, reg
            IMUL reg, mem
            IMUL reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.mul)
        return ''


class Andf(Instruction):
    """
        <instr>
             and
        </instr>
        <syntax>
            AND reg, reg
            AND reg, mem
            AND reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.and_)
        return ''


class Orf(Instruction):
    """
        <instr>
             or
        </instr>
        <syntax>
            OR reg, reg
            OR reg, mem
            OR reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.or_)
        return ''


class Xor(Instruction):
    """
        <instr>
             xor
        </instr>
        <syntax>
            XOR reg, reg
            XOR reg, mem
            XOR reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.xor)
        return ''


class Shl(Instruction):
    """
        <instr>
             shl
        </instr>
        <syntax>
            SHL reg, reg
            SHL reg, mem
            SHL reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.lshift)
        return ''


class Shr(Instruction):
    """
        <instr>
             shr
        </instr>
        <syntax>
            SHR reg, reg
            SHR reg, mem
            SHR reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, line_num, opfunc.rshift)


class Notf(Instruction):
    """
        <instr>
             not
        </instr>
        <syntax>
            NOT reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        one_op_arith(ops, vm, self.name, line_num, opfunc.inv)


class Inc(Instruction):
    """
        <instr>
             inc
        </instr>
        <syntax>
            INC reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        ops[0].set_val(ops[0].get_val(line_num) + 1, line_num)
        vm.changes.add(ops[0].get_nm())


class Dec(Instruction):
    """
        <instr>
             dec
        </instr>
        <syntax>
            DEC reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        ops[0].set_val(ops[0].get_val(line_num) - 1, line_num)
        vm.changes.add(ops[0].get_nm())


class Neg(Instruction):
    """
        <instr>
             neg
        </instr>
        <syntax>
            NEG reg
        </syntax>
        <descr>
           Replaces the value of operand with
           it's two's complement.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        one_op_arith(ops, vm, self.name, line_num, opfunc.neg)
        return ''


class Idiv(Instruction):
    """
        <instr>
             idiv
        </instr>
        <syntax>
            IDIV reg
        </syntax>
        <descr>
            The idiv instruction divides the contents of
            the 64 bit integer EDX:EAX (constructed by viewing
            EDX as the most significant four bytes and EAX
            as the least significant four bytes) by the
            specified operand value. The quotient result
            of the division is stored into EAX, while the
            remainder is placed in EDX.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)

        hireg = int(vm.registers['EDX']) << 32
        lowreg = int(vm.registers['EAX'])
        dividend = hireg + lowreg
        if ops[0].get_val(line_num) == 0:
            raise DivisionZero()
        vm.registers['EAX'] = dividend // ops[0].get_val(line_num)
        vm.registers['EDX'] = dividend % ops[0].get_val(line_num)
        vm.changes.add('EAX')
        vm.changes.add('EDX')
        return ''


class BTR(Instruction):
    """
    <instr>
        btr
    </instr>
    <syntax>
        btr reg, reg
        btr reg, const
    </syntax>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        if ops[1].get_val(line_num).bit_length() >= 9:
            raise InvalidConVal(ops[1].get_nm(), line_num)
        ops[0].set_val(set_bit(ops[0].get_val(line_num),
                               ops[1].get_val(line_num), 0), line_num)


def set_bit(v, index, x):
    mask = 1 << index   # Compute mask, an integer with just bit 'index' set.
    v &= ~mask          # Clear the bit indicated by the mask (if x is False)
    if x:
        v |= mask         # If x was True, set the bit indicated by the mask.
    return v


class BTS(Instruction):
    """
    <instr>
        bts
    </instr>
    <syntax>
        bts reg, reg
        bts reg, const
    </syntax>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        if ops[1].get_val(line_num).bit_length() >= 9:
            raise InvalidConVal(ops[1].get_nm(), line_num)
        ops[0].set_val(set_bit(ops[0].get_val(line_num),
                               ops[1].get_val(line_num), 1), line_num)


class BSF(Instruction):
    """
    <instr>
        bsf
    </instr>
    <syntax>
        bsf reg, reg
        bsf reg, mem
    </syntax>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        num = ops[1].get_val(line_num)

        if num == 0:
            vmachine.flags["ZF"] = 1
        else:
            vmachine.flags["ZF"] = 0
            if ops[1].get_val(line_num).bit_length() <= 16:
                bit_size = 16
            else:
                bit_size = 32
            bin_str = str(bin(num))[2:].zfill(bit_size)[::-1]
            for index in range(len(bin_str)):
                if bin_str[index] == '1':
                    break
            ops[0].set_val(index, line_num)


class BSR(Instruction):
    """
    <instr>
        bsr
    </instr>
    <syntax>
        bsr reg, reg
        bsr reg, mem
    </syntax>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is notof type Register', line_num)
        num = ops[1].get_val(line_num)
        if num == 0:
            vmachine.flags["ZF"] = 1
        else:
            vmachine.flags["ZF"] = 0
            if ops[1].get_val(line_num).bit_length() <= 16:
                bit_size = 16
            else:
                bit_size = 32
            binStr = str(bin(num))[2:].zfill(bit_size)
            for index in range(len(binStr)):
                if binStr[index] == '1':
                    break
            index = bit_size - index
            ops[0].set_val(index, line_num)


class BT(Instruction):
    """
    <instr>
        bt
    </instr>
    <syntax>
        bt reg, reg
        bt reg, const
    </syntax>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        # print('length', ops[1].get_val(line_num).bit_length())
        # if ops[1].get_val(line_num).bit_length() < 8:
        #     raise InvalidOperand('Operand 2 should be of size 8 bit')

        binary_num = bin(ops[0].get_val(line_num))
        index = len(binary_num) - ops[1].get_val(line_num)
        vmachine.flags["CF"] = binary_num[index]


class BTC(Instruction):
    """
    <instr>
        bt
    </instr>
    <syntax>
        bt reg, reg
        bt reg, const
    </syntax>
    """

    def fhook(self, ops, vmachine, line_num):
        check_num_args(self.name, ops, 2, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidOperand('Operand 1 is not of type Register', line_num)
        # if ops[1].get_val(line_num).bit_length() != 8:
        #     raise InvalidOperand('Operand 2 should be of size 8 bit')

        binary_num = bin(ops[0].get_val(line_num))
        index = len(binary_num) - ops[1].get_val(line_num)
        if binary_num[index] == '1':
            complement = '0'
        else:
            complement = '1'
        vmachine.flags["CF"] = complement



"""
lex.py: performs lexical analysis
"""

import re
from .errors import IntOutOfRng, InvalidArgument
from .tokens import Register, NewSymbol, Section
from .tokens import QuestionTok, PlusTok, MinusTok
from .tokens import StringTok, IntegerTok, OpenBracket, CloseBracket
from .tokens import Comma, OpenParen, CloseParen, FloatTok

# for floating point to binary and back
import struct
import binascii

SYM_RE = "([A-Za-z_][A-Za-z0-9_]*)"
sym_match = re.compile(SYM_RE)

FP_RE = "([0-9]+\.[0-9]+)"  # noqa
fp_match = re.compile(FP_RE)

LABEL_RE = SYM_RE + ":"
label_match = re.compile(LABEL_RE)

DATA_SECT = ".data"
TEXT_SECT = ".text"

SEPARATORS = set([',', '(', ')', '[', ']', '+', '-'])

keywords_to_tokens = {
    "[": OpenBracket(),
    "]": CloseBracket(),
    "(": OpenParen(),
    ")": CloseParen(),
    ",": Comma(),
    "+": PlusTok(),
    "-": MinusTok(),
    "?": QuestionTok()
}

# def convert_hex_float(string):
#     lst = string.split(".")
#     int_part = int(lst[0], 16)
#     float_part = float("." + lst[1])
#     return int_part + float_part


def float_to_hex(f):
    return hex(struct.unpack('<I', struct.pack('<f', f))[0])


# to convert the ieee 754 hex back to the actual float value
def hex_to_float(h):
    h2 = h[2:]
    h2 = binascii.unhexlify(h2)
    return struct.unpack('>f', h2)[0]


# for double precision (64 bits) fps
getBin = lambda x: x > 0 and str(bin(x))[2:] or "-" + str(bin(x))[3:]   # noqa


def f_to_b64(value):
    if (value == 0):
        return "0"*64
    val = struct.unpack('q', struct.pack('d', value))[0]
    return getBin(val)


def b_to_f(value):
    if (value == "0"*64):
        return 0.0
    hx = hex(int(value, 2))
    return struct.unpack("d", struct.pack("q", int(hx, 16)))[0]


def generate_reg_dict(vm):
    """
    Generates a dictionary
    Keys: register name
    Values: Register token

    Args:
        vm: Virtual machine

    Returns:
        A dictionary of (registers, register tokens)
    """
    registers = {}
    if vm.flavor != 'wasm':
        for reg in vm.registers:
            if vm.flavor == "att":
                registers["%" + reg] = Register(reg, vm)
            else:
                registers[reg] = Register(reg, vm)
    return registers


def generate_float_stack_dict(vm):
    """
    Generates a dictionary
    Keys: register name
    Values: Register token

    Args:
        vm: Virtual machine

    Returns:
        A dictionary of (registers, register tokens)
    """
    registers = {}

    # for i in range(8):
    #     registers['ST'+str(i)] = NewSymbol('ST'+str(i))
    for reg in vm.fp_stack_registers:
        registers[reg] = Register(reg, vm)

    return registers


def make_language_keys(vm):
    """
    Creates a dictionary of key terms

    Args:
        vm: Virtual machine

    Returns:
        A dictionary of key terms with associated tokens
    """
    language_keys = {}
    language_keys.update(keywords_to_tokens)

    language_keys.update(generate_reg_dict(vm))
    if vm.flavor == "mips_asm" or vm.flavor == "mips_mml":
        from .MIPS.key_words import key_words
        language_keys.update(key_words)
        return language_keys
    elif vm.flavor == "riscv":
        from .RISCV.key_words import key_words
        language_keys.update(key_words)
    elif vm.flavor == 'wasm':
        from .WASM.key_words import key_words
        language_keys.update(key_words)
    else:
        from .Intel.key_words import instructions
        language_keys.update(instructions)
        if vm.flavor == "intel":
            # language_keys.update(generate_float_stack_dict(vm, flavor))
            from .Intel.key_words import intel_key_words
            language_keys.update(intel_key_words)
        else:
            from .Intel.key_words import att_key_words
            language_keys.update(att_key_words)
    return language_keys


def clean_list(lst):
    """
    Removes empty strings from a list of words

    Args:
        lst: List of words

    Returns:
        Cleaned list of words
    """
    while "" in lst:
        lst.remove("")


def split_code(code, vm):
    """
    Splits code on regular expressions and on separators

    Args:
        code: Line of code

    Returns:
        A list of words
    """

    # split on regular expressions
    words = re.split("[ \t\r\n]+", code)
    index = 0

    # split on separator characters, such as commas
    while index < len(words):
        splitter = ""
        for character in words[index]:
            # determine the splitter
            if character in SEPARATORS and words[index] != character:
                splitter = character
                break
            # splitter specifically for AT&T
            elif (vm.flavor == "att" and
                  words[index] != "$" and
                  character == "$"):
                splitter = "$"
                break
        # split if splitter is found
        if splitter != "":
            split_location = words[index].find(splitter)
            temp_words = [words[index][:split_location]]
            temp_words.append(splitter)
            temp_words.append(words[index][split_location + 1:])
            words = words[:index] + temp_words + words[index + 1:]
        else:
            index += 1

    # remove all empty strings from list that were made from splitting
    # if splitter was first or last character
    clean_list(words)
    return words


def sep_line(code, line_num, i, data_sec, vm, language_keys):
    """
    Returns a list of tokens created

    Args:
        code: Line of code
        line_num: Line number of code, including the data section
        i: Line number of code
           Needed for determining label location
        data_sec: Boolean, determines if we are in the data section
                  Needed to differentiate between label and symbol
        vm: Virtual machine
        key_words: Dictionary of key words for the flavor

    Returns:
        Tuple of the lexical analysis of the line
        The first member is the tokens and the second is the
        text of the code.
    """
    analysis = []
    words = split_code(code, vm)
    # for i in range(len(words)):  #fixes parsing error with negative floats
    #     if words[i]=='-':
    #         words[i+1]='-'+words[i+1]
    # words = [x for x in words if x!='-']
    data_type = None
    for word in words:
        # keyword:
        if word.upper() in language_keys:
            analysis.append(language_keys[word.upper()])
            data_type = word
        # section declaration:
        elif word[0] == ".":
            analysis.append(Section(word[1:]))
        # string:
        elif word.find("'") != -1:
            analysis.append(StringTok(word))
        # label / symbol:
        elif re.match(label_match, word) is not None:
            if vm.flavor == "intel":
                vm.labels[word[:-1]] = i
            else:
                if data_sec:
                    analysis.append(NewSymbol(word[:-1], vm))
                else:
                    if vm.flavor == "mips_asm" or vm.flavor == "riscv":
                        vm.labels[word[:-1]] = i * 4
                    else:
                        vm.labels[word[:-1]] = i
        elif re.match(sym_match, word) is not None:
            analysis.append(NewSymbol(word, vm))
        # Floating Points
        elif re.match(fp_match, word) is not None:
            # default is float (single precision) if user doesnt say
            if (data_type != ".float" and data_type != ".double" and
                    data_type != "REAL4" and data_type != "REAL8"):
                data_type = ".float"
            if vm.base == "dec":
                # TODO: Screen shot to give me the
                # floating point token class from token.py
                analysis.append(FloatTok(data_type=data_type, val=float(word)))
            else:   # hexadecimal

                if data_type == ".float":

                    analysis.append(FloatTok(data_type=data_type,
                                             val=float_to_hex(float(word))))
                elif data_type == ".double":
                    analysis.append(FloatTok(data_type=data_type,
                                             val=f_to_b64(float(word))))
        # Integers
        else:
            if vm.base == "dec":
                try:
                    if vm.flavor == "att":
                        analysis.append(IntegerTok(int(word), False))
                    else:
                        analysis.append(IntegerTok(int(word)))
                except IntOutOfRng:
                    raise IntOutOfRng(word, line_num)
                except Exception:
                    raise InvalidArgument(word, line_num)
            else:
                try:
                    if vm.flavor == "att":
                        analysis.append(IntegerTok(int(word, 16), False))
                    else:
                        analysis.append(IntegerTok(int(word, 16)))
                except IntOutOfRng:
                    raise IntOutOfRng(word, line_num)
                except Exception:
                    raise InvalidArgument(word, line_num)
    return (analysis, code, line_num)


def sep_line_mml(code, line_num, i, vm, language_keys):
    """
    Returns a list of tokens created

    Args:
        code: Line of code
        line_num: Line number of code, including the data section
        i: Line number of code
           Needed for determining label location
        data_sec: Boolean, determines if we are in the data section
                  Needed to differentiate between label and symbol
        vm: Virtual machine
        key_words: Dictionary of key words for the flavor

    Returns:
        Tuple of the lexical analysis of the line
        The first member is the tokens and the second is the
        text of the code.
    """
    analysis = []
    words = split_code(code, vm)

    for word in words:
        # keyword
        if word.upper() in language_keys:
            analysis.append(language_keys[word.upper()])
    # Integers
        else:
            try:
                analysis.append(IntegerTok(int(word, 16)))
            except IntOutOfRng:
                raise IntOutOfRng(word, line_num)
            except Exception:
                raise InvalidArgument(word, line_num)
    return (analysis, code, line_num)


def sep_line_wasm(code, line_num, i, vm, language_keys):
    """
    Returns a list of tokens created

    Args:
        code: Line of code
        line_num: Line number of code, including the data section
        i: Line number of text code
           Needed for determining label location
        data_sec: Boolean, determines if we are in the data section
                  Needed to differentiate between label and symbol
        vm: Virtual machine
        key_words: Dictionary of key words for the flavor

    Returns:
        Tuple of the lexical analysis of the line
        The first member is the tokens and the second is the
        text of the code.
    """
    analysis = []
    words = split_code(code, vm)

    for word in words:
        # keyword
        if word in language_keys:
            analysis.append(language_keys[word])
        # variable symbol
        elif re.match(sym_match, word) is not None:
            analysis.append(NewSymbol(word, vm))
    # Integers
        else:
            try:
                analysis.append(IntegerTok(int(word)))
            except IntOutOfRng:
                raise IntOutOfRng(word, line_num)
            except Exception:
                raise InvalidArgument(word, line_num)
    return (analysis, code, line_num)


def lex(code, vm):
    """
    Lexical phase: tokenizes the code.

    Args:
        code: The code to lexically analyze.
        vm: virtual machine

    Returns:
        tok_lines: the tokenized version
    """
    lines = code.split("\n")

    pre_processed_lines = []
    tok_lines = []  # this will hold the tokenized version of the code
    i = 0
    add_to_ip = True
    data_sec = False    # used for AT&T version
    for index in range(len(lines)):
        line = lines[index]
        # comments:
        comm_start = line.find(";")
        if comm_start > 0:  # -1 means not found
            line = line[0:comm_start]
        elif comm_start == 0:  # the whole line is a comment
            continue

        # strip AFTER comments to handle blanks between code and ;
        line = line.strip()
        if len(line) == 0:  # blank lines ok; just skip 'em
            continue

        # index starts at 0, line numbers start at 1
        pre_processed_lines.append([line, index + 1])

    # we've stripped extra whitespace, comments, and labels:
    # now perform lexical analysis
    for line, line_num in pre_processed_lines:
        # create language-specific dictionary:
        language_keys = make_language_keys(vm)
        if vm.flavor == "mips_mml":
            tok_lines.append(sep_line_mml(line, line_num, i,
                                          vm, language_keys))
        elif vm.flavor == "wasm":
            tok_lines.append(sep_line_wasm(line, line_num, i,
                                           vm, language_keys))
        else:
            tok_lines.append(sep_line(line, line_num, i, data_sec,
                                      vm, language_keys))
        if line == ".data":
            add_to_ip = False
            data_sec = True
            continue
        if line == ".text":
            add_to_ip = True
            data_sec = False
            continue

        # we count line numbers to store label jump locations:
        if add_to_ip:
            i += 1
    return tok_lines
